"""Module contains Label model."""

from crux.models.model import CruxModel


class Label(CruxModel):
    """Label Model."""

    @property
    def label_key(self):
        """str: Gets the Label Key."""
        return self.raw_model["labelKey"]

    @property
    def label_value(self):
        """str: Gets the Label Value."""
        return self.raw_model["labelValue"]



"""Module contains Resource model."""

from enum import Enum
import os
import posixpath
from typing import Dict, List, Union  # noqa: F401

from requests.models import Response  # noqa: F401 pylint: disable=unused-import

from crux._client import CruxClient
from crux._utils import create_logger, DEFAULT_CHUNK_SIZE, Headers
from crux.models.model import CruxModel
from crux.models.permission import Permission


log = create_logger(__name__)


class Resource(CruxModel):
    """Resource Model."""

    def __init__(self, raw_model=None, connection=None):
        # type: (Dict, CruxClient) -> None
        """
        Attributes:
            raw_model (dict): Resource raw dictionary. Defaults to None.
            connection (CruxClient): Connection Object. Defaults to None.
        """
        self._folder = None
        super(Resource, self).__init__(raw_model, connection)

    @property
    def id(self):
        """str: Gets the Resource ID."""
        return self.raw_model["resourceId"]

    @property
    def description(self):
        """str: Gets the Resource Description."""
        return self.raw_model["description"]

    @description.setter
    def description(self, description):
        self.raw_model["description"] = description

    @property
    def media_type(self):
        """str: Gets the Media type."""
        return self.raw_model["mediaType"]

    @property
    def dataset_id(self):
        """str: Gets the Dataset ID."""
        return self.raw_model["datasetId"]

    @property
    def folder_id(self):
        """str: Gets the Folder ID."""
        return self.raw_model["folderId"]

    @folder_id.setter
    def folder_id(self, folder_id):
        self.raw_model["folderId"] = folder_id

    @property
    def frame_id(self):
        """str: Gets the Frame ID."""
        return self.labels["frame_id"]

    @property
    def storage_id(self):
        """str: Gets the Storage ID."""
        return self.raw_model["storageId"]

    @property
    def name(self):
        """str: Gets the Resource Name."""
        return self.raw_model["name"]

    @name.setter
    def name(self, name):
        self.raw_model["name"] = name

    @property
    def provenance(self):
        """dict: Gets the Provenance."""
        return self.raw_model["provenance"]

    @provenance.setter
    def provenance(self, provenance):
        self.raw_model["provenance"] = provenance

    @property
    def supplier_implied_dt(self):
        """str: Gets the supplier date."""
        return self.labels["supplier_implied_dt"]

    @property
    def type(self):
        """str: Gets the Resource Type."""
        return self.raw_model["type"]

    @property
    def tags(self):
        """:obj:`list` of :obj:`str`: Gets the Resource Tags."""
        return self.raw_model["tags"]

    @tags.setter
    def tags(self, tags):
        self.raw_model["tags"] = tags

    @property
    def labels(self):
        """dict: Gets the Resource labels."""
        labels_dict = {}
        for label in self.raw_model["labels"]:
            labels_dict[label["labelKey"]] = label["labelValue"]
        return labels_dict

    @property
    def as_of(self):
        """str: Gets the as_of."""
        return self.raw_model["asOf"]

    @property
    def created_at(self):
        """str: Gets created_at."""
        return self.raw_model["createdAt"]

    @property
    def ingestion_time(self):
        """str: Gets created_at."""
        return self.labels["ingestion_dt"]

    @property
    def modified_at(self):
        """str: Gets modified_at."""
        return self.raw_model["modifiedAt"]

    @property
    def size(self):
        """int: Gets the size."""
        return self.raw_model["size"]

    @property
    def path(self):
        """str: Compute or Get the resource path."""
        return posixpath.join(self.folder, self.name)

    @property
    def folder(self):
        """str: Compute or Get the folder name."""
        if self._folder:
            return self._folder

        self._folder = self._get_folder()
        return self._folder

    def delete(self):
        # type: () -> bool
        """Deletes Resource from Dataset.

        Returns:
            bool: True if it is deleted.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        return self.connection.api_call("DELETE", ["resources", self.id], headers=headers)

    def update(self, name=None, description=None, tags=None, provenance=None):
        # type: (str, str, List[str], str) -> bool
        """Updates the metadata for Resource.

        Args:
            name (str): Name of resource. Defaults to None.
            description (str): Description of the resource. Defaults to None.
            tags (:obj:`list` of :obj:`str`): List of tags. Defaults to None.
            provenance (str): Provenance for a resource. Defaults to None.

        Returns:
            bool: True, if resource is updated.

        Raises:
            ValueError: It is raised if name, description or tags are unset.
            TypeError: It is raised if tags are not of type List.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        body = {}  # type: Dict[str, Union[str, List, Dict]]

        if name is not None:
            self.raw_model["name"] = name
        if description is not None:
            self.raw_model["description"] = description
        if tags is not None:
            self.raw_model["tags"] = tags
        if provenance is not None:
            self.raw_model["provenance"] = provenance

        body = self.raw_model

        log.debug("Body %s", body)

        resource_object = self.connection.api_call(
            "PUT", ["resources", self.id], headers=headers, json=body, model=Resource
        )

        self.raw_model = resource_object.raw_model

        log.debug("Updated dataset %s with content %s", self.id, self.raw_model)
        return True

    def add_permission(self, identity_id, permission):
        # type: (str, str) -> Union[bool, Permission]
        """Adds permission to the resource.

        Args:
            identity_id: Identity Id to be set.
            permission: Permission to be set.

        Returns:
            crux.models.Permission: Permission Object.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        return self.connection.api_call(
            "PUT",
            ["permissions", self.id, identity_id, permission],
            model=Permission,
            headers=headers,
        )

    def delete_permission(self, identity_id, permission):
        # type: (str, str) -> bool
        """Deletes permission from the resource.

        Args:
            identity_id (str): Identity Id for the deletion.
            permission (str): Permission for the deletion.

        Returns:
            bool: True if it is able to delete it.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        return self.connection.api_call(
            "DELETE", ["permissions", self.id, identity_id, permission], headers=headers
        )

    def list_permissions(self):
        # type: () -> List[Permission]
        """Lists the permission on the resource.

        Returns:
            list (:obj:`crux.models.Permission`): List of Permission Objects.
        """
        headers = Headers({"accept": "application/json"})
        return self.connection.api_call(
            "GET", ["resources", self.id, "permissions"], model=Permission, headers=headers,
        )

    def add_label(self, label_key, label_value):
        # type: (str, str) -> bool
        """Adds label to Resource.

        Args:
            label_key (str): Label Key for Resource.
            label_value (str): Label Value for Resource.

        Returns:
            bool: True if label is added, False otherwise.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        response_result = self.connection.api_call(
            "PUT",
            [
                "datasets",
                self.dataset_id,
                "resources",
                self.id,
                "labels",
                label_key,
                label_value,
            ],
            headers=headers,
        )

        if response_result:
            # Sync the latest data from API to prevent inconsistency
            self.refresh()

        return True

    def delete_label(self, label_key):
        # type: (str) -> bool
        """Deletes label from Resource.

        Args:
            label_key (str): Label Key for Resource.

        Returns:
            bool: True if label is deleted, False otherwise.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        response_result = self.connection.api_call(
            "DELETE",
            ["datasets", self.dataset_id, "resources", self.id, "labels", label_key],
            headers=headers,
        )

        if response_result:
            # Sync the latest data from API to prevent inconsistency
            self.refresh()

        return True

    def add_labels(self, labels_dict):
        # type: (dict) -> bool
        """Adds multiple labels to Resource.

        Args:
            label_dict (dict): Labels (key/value pairs) to add to the Resource.

        Returns:
            bool: True if the labels were added, False otherwise.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})

        labels_list = []
        for label_key, label_value in labels_dict.items():
            if label_key is not None and label_value is not None:
                label_key = label_key.value if isinstance(label_key, Enum) else label_key
                labels_list.append(
                    {"labelKey": str(label_key), "labelValue": str(label_value)}
                )

        data = {"labels": labels_list}

        response_result = self.connection.api_call(
            "PUT",
            ["datasets", self.dataset_id, "resources", self.id, "labels"],
            headers=headers,
            json=data,
        )

        if response_result:
            # Sync the latest data from API to prevent inconsistency
            self.refresh()

        return True

    def _get_folder(self):
        # type: () -> str
        """Fetches the folder of the resource.

        Returns:
            str: Folder name of the resource.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        response = self.connection.api_call(
            "GET", ["resources", self.id, "folderpath"], headers=headers
        )

        return response.json().get("path")

    def _download(self, file_obj, media_type, chunk_size=DEFAULT_CHUNK_SIZE):

        if media_type is not None:
            headers = Headers({"accept": media_type})
        else:
            headers = None

        data = self.connection.api_call(
            "GET", ["resources", self.id, "content"], headers=headers, stream=True
        )

        for chunk in data.iter_content(chunk_size=chunk_size):
            file_obj.write(chunk)
        data.close()
        return True

    def refresh(self):
        """Refresh Resource model from API backend.

        Returns:
            bool: True, if it is able to refresh the model,
                False otherwise.
        """
        # type () -> bool
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        resource_object = self.connection.api_call(
            "GET", ["resources", self.id], headers=headers, model=Resource
        )

        self.raw_model = resource_object.raw_model

        return True


# https://github.com/python/mypy/issues/2477, mypy is performing checking with Python2
class MediaType(Enum):  # type: ignore
    """MediaType Enumeration Model."""

    JSON = "application/json"
    NDJSON = "application/x-ndjson"
    CSV = "text/csv"
    PARQUET = "parquet/binary"
    AVRO = "avro/binary"

    @classmethod
    def detect(cls, file_name):
        # type: (str) -> str
        """Detects the media_type from the file extension.

        Args:
            file_name (str): Absolute or Relative Path of the file.

        Returns:
            str: MediaType extension.

        Raises:
            LookupError: If file type is not supported.
        """
        file_ext = os.path.splitext(file_name)[1][1:].upper()

        if file_ext in cls.__members__:
            return cls[file_ext].value  # type: ignore
        else:
            raise LookupError("File/Media Type not supported.")



"""Module handles import compatibility issues between Python 2 and Python 3."""
# pylint: disable=useless-suppression,ungrouped-imports

try:
    # Python 3 imports
    from builtins import str as unicode
    from urllib.parse import quote as urllib_quote  # type: ignore
except ImportError:
    # Python 2 imports
    from __builtin__ import unicode  # type: ignore
    from urllib import quote as urllib_quote

__all__ = ("unicode", "urllib_quote")



"""Module contains Identity model."""

from crux.models.model import CruxModel


class Identity(CruxModel):
    """Identity Model."""

    @property
    def identity_id(self):
        """str: Gets the Identity Id."""
        return self.raw_model["identityId"]

    @property
    def parent_identity_id(self):
        """str: Gets the Parent Identity Id."""
        return self.raw_model["parentIdentityId"]

    @property
    def company_name(self):
        """str: Gets the Company name."""
        return self.raw_model["companyName"]

    @company_name.setter
    def company_name(self, company_name):
        self.raw_model["companyName"] = company_name

    @property
    def description(self):
        """str: Gets the Description."""
        return self.raw_model["description"]

    @description.setter
    def description(self, description):
        self.raw_model["description"] = description

    @property
    def first_name(self):
        """str: Gets the First name."""
        return self.raw_model["firstName"]

    @first_name.setter
    def first_name(self, first_name):
        self.raw_model["firstName"] = first_name

    @property
    def last_name(self):
        """str: Gets the Last name."""
        return self.raw_model["lastName"]

    @last_name.setter
    def last_name(self, last_name):
        self.raw_model["lastName"] = last_name

    @property
    def role(self):
        """str: Gets the Role."""
        return self.raw_model["role"]

    @role.setter
    def role(self, role):
        self.raw_model["role"] = role

    @property
    def email(self):
        """str: Gets the Email."""
        return self.raw_model["email"]

    @email.setter
    def email(self, email):
        self.raw_model["email"] = email

    @property
    def website(self):
        """str: Gets the Website."""
        return self.raw_model["website"]

    @website.setter
    def website(self, website):
        self.raw_model["website"] = website

    @property
    def landing_page(self):
        """str: Gets the Landing Page."""
        return self.raw_model["landingPage"]

    @landing_page.setter
    def landing_page(self, landing_page):
        self.raw_model["landingPage"] = landing_page

    @property
    def type(self):
        """str: Gets the Type."""
        return self.raw_model["type"]

    @property
    def phone(self):
        """str: Gets the phone."""
        return self.raw_model["phone"]

    @phone.setter
    def phone(self, phone):
        self.raw_model["phone"] = phone



"""Module provides CruxConfig object to manage API configuration settings."""

import os
import platform
import re
from typing import Dict, MutableMapping, Optional, Text, Union  # noqa: F401

import requests
from requests.packages.urllib3.util.retry import (  # Dynamic load pylint: disable=import-error
    Retry,
)

from crux.__version__ import __version__
from crux._utils import create_logger, get_session, str_to_bool

log = create_logger(__name__)


class CruxConfig(object):
    """
    Crux Configuration Class.
    """

    def __init__(  # pylint: disable=too-many-branches
        self,
        api_key=None,  # type: Optional[str]
        api_host=None,  # type: str
        api_prefix=None,  # type: str
        proxies=None,  # type: Optional[MutableMapping[Text, Text]]
        user_agent=None,  # type: str
        only_use_crux_domains=None,  # type: bool
        session=None,  # type: requests.Session
    ):
        # type: (...) -> None
        """
        Args:
            api_key (str): API Key. Defaults to None.
            api_host (str): API URL. Defaults to None.
            api_prefix (str): API prefix to be used. Defaults to None.
            proxies (dict): Proxies to be used. Defaults to None.
            user_agent (str): User agent to be used. Defaults to None.
            only_use_crux_domains (bool): True if Crux domain should be
                use for upload and download, False otherwise.
                Defaults to False.
            session(requests.Session): Session to be used with connection.

        Raises:
            ValueError: If CRUX_API_KEY is not set.
        """
        if api_key is None:
            if "CRUX_API_KEY" in os.environ:
                log.debug("Fetching API KEY from OS Environment Variable")
                self.api_key = os.environ.get("CRUX_API_KEY")  # type: Optional[str]
                log.trace("API KEY: %s", self.api_key)
            else:
                raise ValueError("API KEY is required")
        else:
            self.api_key = api_key
            log.trace("API KEY: %s", self.api_key)

        if api_host is None:
            self.api_host = os.environ.get(
                "CRUX_API_HOST", "https://api.cruxinformatics.com"
            )
            log.debug("Setting API host to %s", self.api_host)
        else:
            self.api_host = api_host
            log.debug("Setting API host to %s", self.api_host)

        if api_prefix is None:
            self.api_prefix = os.environ.get("CRUX_API_PREFIX", "plat-api")
            log.trace("Setting API prefix to %s", self.api_prefix)
        else:
            self.api_prefix = api_prefix
            log.trace("Setting API prefix to %s", self.api_prefix)

        if user_agent is None:
            self.user_agent = self._default_user_agent()
            log.debug("Setting User Agent to %s", self.user_agent)
        else:
            self.user_agent = user_agent
            log.debug("Setting User Agent to %s", self.user_agent)

        self.proxies = (
            proxies if proxies else {}
        )  # type: Optional[MutableMapping[Text, Text]]

        if only_use_crux_domains is None:
            self.only_use_crux_domains = str_to_bool(
                os.environ.get("CRUX_ONLY_USE_CRUX_DOMAINS", "false")
            )
            log.debug("Setting only_use_crux_domain to %s", self.only_use_crux_domains)
        else:
            self.only_use_crux_domains = only_use_crux_domains
            log.debug("Setting only_use_crux_domain to %s", self.only_use_crux_domains)

        if session is None:
            retries = Retry(
                total=20,
                backoff_factor=0.3,
                status_forcelist=(
                    500,
                    502,
                    503,
                    504,
                    520,
                    521,
                    522,
                    523,
                    524,
                    525,
                    527,
                    530,
                ),
                method_whitelist=("GET", "PUT", "DELETE", "POST"),
                redirect=10,
                connect=10,
                read=10,
            )
            self.session = get_session(retries=retries, proxies=self.proxies)
        else:
            self.session = session

    def _default_user_agent(self):
        # type: () -> str
        user_agent = (
            "crux-python/{ver}"
            " requests/{req_ver} {py_impl}/{py_ver} "
            "{os}/{os_ver} {cpu}/{machine}"
        ).format(
            ver=__version__,
            req_ver=self._sanitize_user_agent_part(requests.__version__),
            py_impl=self._sanitize_user_agent_part(platform.python_implementation()),
            py_ver=self._sanitize_user_agent_part(platform.python_version()),
            os=self._sanitize_user_agent_part(platform.system()),
            os_ver=self._sanitize_user_agent_part(platform.release()),
            cpu=self._sanitize_user_agent_part(platform.processor()),
            machine=self._sanitize_user_agent_part(platform.machine()),
        )

        return user_agent

    def _sanitize_user_agent_part(self, part):  # pylint: disable=no-self-use
        # type: (str) -> str
        if part:
            no_space_part = re.sub(r"\s+", "_", part)
            sanitized_part = re.sub(r"[^a-zA-Z0-9._+~-]", "", no_space_part)
            if sanitized_part:
                return sanitized_part
        return "unknown"



"""Module defines abstract CruxModel."""

import copy
import pprint
from typing import Any, Dict  # noqa: F401

from crux._client import CruxClient
from crux._client import CruxConfig


class CruxModel(object):
    """Base Crux model."""

    def __init__(self, raw_model=None, connection=None):
        # type: (Dict, CruxClient) -> None
        """
        Attributes:
            raw_model (dict): Resource raw dictionary. Defaults to None.
            connection (CruxClient): Connection object. Defaults to None.
        """
        self.raw_model = raw_model if raw_model is not None else {}
        self._connection = connection

    @property
    def connection(self):
        """CruxClient: API connection client."""
        if self._connection is None:
            self._connection = CruxClient(CruxConfig())

        return self._connection

    @connection.setter
    def connection(self, connection):
        self._connection = connection

    def to_dict(self):
        # type: () -> Dict[str, Any]
        """Returns dict copy of raw model.

        Returns:
            dict: Raw model dict.
        """

        # Deep copy raw_model, otherwise if a user modifies the returned dict,
        # they will be mutating this instances raw_model.
        return copy.deepcopy(self.raw_model)

    @classmethod
    def from_dict(cls, a_dict, connection=None):
        # type: (Dict[str, Any], CruxClient) -> Any
        """Returns model instance created from raw model dict.

        Args:
            a_dict (dict): Model dict.
            connection (CruxClient): Connection bbject. Defaults to None.

        Returns:
            crux.models.model.CruxModel: Model instance.
        """

        return cls(raw_model=a_dict, connection=connection)

    def to_str(self):
        # type: () -> str
        """Absract to_str method."""
        return pprint.pformat(self.raw_model)

    def __repr__(self):
        # type: () -> str
        """Absract __repr__ method."""
        return self.to_str()



"""Module contains Delivery model."""

from typing import Dict, Iterator

from crux._client import CruxClient
from crux.models.file import File
from crux.models.model import CruxModel
from crux.models.resource import MediaType, Resource


class Delivery(CruxModel):
    """Delivery Model."""

    def __init__(self, raw_model=None, connection=None):
        # type: (Dict, CruxClient) -> None
        """
        Attributes:
            raw_model (dict): Ingestion raw dictionary, Defaults to None.
            connection (CruxClient): Connection Object. Defaults to None.
        """
        self._summary = None
        super(Delivery, self).__init__(raw_model, connection)

    @property
    def id(self):
        """str: Gets the Delivery ID."""
        return self.raw_model["delivery_id"]

    @property
    def dataset_id(self):
        """str: Gets the Dataset ID."""
        return self.raw_model["dataset_id"]

    @property
    def ingestion_time(self):
        """str: Gets ingestion time of delivery."""
        return self.summary["ingestion_time"]

    @property
    def status(self):
        """str: Gets the Status of delivery."""
        return self.summary["latest_health_status"]

    @property
    def schedule_datetime(self):
        """str: Gets schedule datetime of delivery."""
        return self.summary["schedule_dt"]

    @property
    def summary(self):
        """dict: Gets the Delivery Summary"""
        if self._summary is None:
            response = self.connection.api_call(
                "GET", ["deliveries", self.dataset_id, self.id]
            )
            self._summary = response.json()
        return self._summary

    def get_data(self, file_format=MediaType.AVRO.value, use_cache=None):
        # type: (str, bool) -> Iterator[Resource]
        """Get the processed delivery data

        Args:
            file_format (str): File format of delivery.
            use_cache (bool): Preference to set cached response

        Returns:
            list (:obj:`crux.models.Resource`): List of resources.
        """
        params = {}
        params["delivery_resource_format"] = file_format
        if use_cache is not None:
            params["useCache"] = use_cache

        response = self.connection.api_call(
            "GET", ["deliveries", self.dataset_id, self.id, "data"], params=params
        )

        resource_list = response.json()["resources"]

        if resource_list:
            for resource in resource_list:
                obj = File(raw_model={"resourceId": resource["resource_id"]})
                obj.connection = self.connection
                obj.refresh()
                yield obj

    def get_raw(self, use_cache=None):
        # type: (bool) -> Iterator[Resource]
        """Get the raw delivery data

        Args:
            use_cache (bool): Preference to set cached response

        Returns:
            list (:obj:`crux.models.Resource`): List of resources.
        """
        params = {}
        if use_cache is not None:
            params["useCache"] = use_cache

        response = self.connection.api_call(
            "GET", ["deliveries", self.dataset_id, self.id, "raw"], params=params
        )

        resource_list = response.json()["resource_ids"]

        if resource_list:
            for resource in resource_list:
                obj = File(
                    raw_model={"resourceId": resource}, connection=self.connection
                )
                obj.refresh()
                yield obj

    def get_healthlog(self, use_cache=None):
        # type: (bool) -> Iterator[Resource]
        """Get delivery healthlog information

        Args:
            use_cache (bool): Preference to set cached response

        Returns:
            dict: Healthlog Json Object.
        """
        params = {}
        if use_cache is not None:
            params["useCache"] = use_cache

        response = self.connection.api_call(
            "GET", ["deliveries", self.id, "log"], params=params
        )

        healthlog_list = response.json()
        return healthlog_list


"""Module contains File model."""

from typing import Any, Dict, Union  # noqa: F401

from crux._utils import create_logger, Headers
from crux.models.permission import Permission
from crux.models.resource import Resource


log = create_logger(__name__)


class Folder(Resource):
    """Folder Model."""

    def add_permission(  # It is by design pylint: disable=arguments-differ
        self, identity_id, permission, recursive=False
    ):
        # type: (str, str, bool) -> Union[bool, Permission]
        """Adds permission to the Folder resource.

        Args:
            identity_id (str): Identity Id to be set.
            permission (str): Permission to be set.
            recursive (bool): If recursive is set to True, it will recursive apply
                permission to all resources under the folder resource.

        Returns:
            bool or crux.models.Permission: If recursive is set then it returns True.
                If recursive is unset then it returns Permission object.
        """

        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        body = {
            "identityId": identity_id,
            "permission": permission,
            "action": "add",
            "resourceIds": [self.id],
        }
        if recursive:
            log.debug(
                "Adding permission %s for %s in recursive mode to resource %s",
                permission,
                identity_id,
                self.id,
            )
            return self.connection.api_call(
                "POST", ["permissions", "bulk"], headers=headers, json=body
            )
        else:
            log.debug(
                "Adding permission %s for %s to resource %s",
                permission,
                identity_id,
                self.id,
            )
            return self.connection.api_call(
                "PUT",
                ["permissions", self.id, identity_id, permission],
                model=Permission,
                headers=headers,
            )

    def delete_permission(  # It is by design pylint: disable=arguments-differ
        self, identity_id, permission, recursive=False
    ):
        # type: (str, str, bool) -> bool
        """Deletes permission from Folder resource.

        Args:
            identity_id (str): Identity Id for the deletion.
            permission (str): Permission for deletion.
            recursive (bool): If recursive is set to True, it will recursively delete
                permission from all resources under the folder resource.
                Defaults to False.

        Returns:
            bool: True if it is able to delete it.
        """
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )

        body = {
            "identityId": identity_id,
            "permission": permission,
            "action": "delete",
            "resourceIds": [self.id],
        }
        if recursive:
            log.debug(
                "Deleting permission %s for %s in recursive mode from resource %s",
                permission,
                identity_id,
                self.id,
            )
            return self.connection.api_call(
                "POST", ["permissions", "bulk"], headers=headers, json=body
            )
        else:
            log.debug(
                "Deleting permission %s for %s in from resource %s",
                permission,
                identity_id,
                self.id,
            )
            return self.connection.api_call(
                "DELETE",
                ["permissions", self.id, identity_id, permission],
                headers=headers,
            )



"""Module contains Permission model."""

from crux.models.model import CruxModel


class Permission(CruxModel):
    """Permission Model."""

    @property
    def target_id(self):
        """str: Gets the Target ID."""
        return self.raw_model["targetId"]

    @property
    def identity_id(self):
        """str: Gets the Identity ID."""
        return self.raw_model["identityId"]

    @property
    def permission_name(self):
        """str: Gets the Permission Name."""
        return self.raw_model["permissionName"]



"""Module contains Ingestion model."""

from typing import Dict, Iterator

from crux._client import CruxClient
from crux._utils import create_logger
from crux.models.delivery import Delivery
from crux.models.model import CruxModel
from crux.models.resource import MediaType, Resource

log = create_logger(__name__)


class Ingestion(CruxModel):
    """Ingestion Model."""

    def __init__(self, raw_model=None, connection=None):
        # type: (Dict, CruxClient) -> None
        """
        Attributes:
            raw_model (dict): Ingestion raw dictionary, Defaults to None.
            connection (CruxClient): Connection Object. Defaults to None.
        """
        self._delivery_objects = {}  # type: Dict[int, Delivery]
        super(Ingestion, self).__init__(raw_model, connection)

    @property
    def id(self):
        """str: Gets the Ingestion ID."""
        return self.raw_model["ingestionId"]

    @property
    def dataset_id(self):
        """str: Gets the Dataset ID."""
        return self.raw_model["datasetId"]

    @property
    def versions(self):
        """list: Gets the list of versions."""
        return sorted(self.raw_model["versions"])

    def _get_delivery_object(self, version=None):
        if version not in self._delivery_objects:
            delivery_id = "{}.{}".format(self.id, version)
            delivery_object = Delivery.from_dict(
                {"delivery_id": delivery_id, "dataset_id": self.dataset_id},
                connection=self.connection,
            )
            self._delivery_objects[version] = delivery_object
        return self._delivery_objects[version]

    def get_data(
        self,
        version=None,  # type: int
        file_format=MediaType.AVRO.value,  # type: str
    ):
        # type: (...) -> Iterator[Resource]
        """Get the processed delivery data

        Args:
            version (int): Version of the delivery.
            file_format (str): File format of delivery.
            accepted status (:obj:`list` of :obj:`str`): List of acceptable statuses.
                Defaults to None.

        Returns:
            list (:obj:`crux.models.Resource`): List of resources.
        """
        if version is None:
            for version_no in sorted(self.versions, reverse=True):
                delivery_object = self._get_delivery_object(version_no)
                if delivery_object.status == "DELIVERY_SUCCEEDED":
                    return delivery_object.get_data(file_format=file_format)
            log.info(
                "Delivery %s has no version with DELIVERY_SUCCEEDED status",
                delivery_object.id,
            )
            return iter([])

        delivery_object = self._get_delivery_object(version)
        return delivery_object.get_data(file_format=file_format)

    def get_raw(self, version=None):
        # type: (...) -> Iterator[Resource]
        """Get the raw delivery data

        Args:
            version (int): Version of the delivery.

        Returns:
            list (:obj:`crux.models.Resource`): List of resources.
        """
        if version is None:
            version = max(self.versions)

        delivery_id = "{}.{}".format(self.id, version)

        delivery_object = Delivery.from_dict(
            {"delivery_id": delivery_id, "dataset_id": self.dataset_id}
        )
        delivery_object.connection = self.connection

        return delivery_object.get_raw()



"""Module contains Dataset model."""

from collections import defaultdict
from datetime import datetime, timedelta
from dateutil import parser
import json
import os
import posixpath
from typing import (
    DefaultDict,
    Dict,
    Generator,
    IO,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Set,
    Text,
    Tuple,
    Union,
)  # noqa: F401

from crux._compat import unicode
from crux._utils import (
    create_logger,
    DELIVERY_ID_REGEX,
    Headers,
    split_posixpath_filename_dirpath,
)
from crux.exceptions import CruxAPIError, CruxClientError, CruxResourceNotFoundError
from crux.models._factory import get_resource_object
from crux.models.delivery import Delivery
from crux.models.file import File
from crux.models.folder import Folder
from crux.models.ingestion import Ingestion
from crux.models.job import StitchJob
from crux.models.label import Label
from crux.models.model import CruxModel
from crux.models.permission import Permission
from crux.models.resource import Resource
from crux.models.resource import MediaType


log = create_logger(__name__)


class Dataset(CruxModel):
    """Dataset Model."""

    @property
    def id(self):
        """str: Gets the Dataset ID."""
        return self.raw_model["datasetId"]

    @property
    def owner_identity_id(self):
        """str: Gets the Owner Identity ID."""
        return self.raw_model.get("ownerIdentityId", "")

    @property
    def contact_identity_id(self):
        """str: Gets the Contact Identity ID."""
        return self.raw_model.get("contactIdentityId", "")

    @property
    def name(self):
        """str: Gets the Dataset Name."""
        return self.raw_model["name"]

    @name.setter
    def name(self, name):
        self.raw_model["name"] = name

    @property
    def tags(self):
        """str: Gets the tags.

        Raises:
            TypeError: If tags is not a list
        """
        return self.raw_model.get("tags", [])

    @tags.setter
    def tags(self, tags):
        if not isinstance(tags, list):
            raise TypeError("Tags should be of type list")
        self.raw_model["tags"] = tags

    @property
    def description(self):
        """str: Gets the Dataset Description."""
        return self.raw_model.get("description", "")

    @description.setter
    def description(self, description):
        self.raw_model["description"] = description

    @property
    def website(self):
        """str: Gets the Dataset Website."""
        return self.raw_model.get("website", "")

    @website.setter
    def website(self, website):
        self.raw_model["website"] = website

    @property
    def created_at(self):
        """str: Gets the Dataset created_at."""
        return self.raw_model.get("createdAt", "")

    @property
    def modified_at(self):
        """str: Gets the Dataset modified_at."""
        return self.raw_model.get("modifiedAt", "")

    @property
    def provenance(self):
        """str: Compute or Get the provenance."""
        return self.raw_model.get("provenance", "")

    def create(self):
        # type: () -> bool
        """Creates the Dataset.

        Returns:
            bool: True if dataset is created.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        dataset_object = self.connection.api_call(
            "POST", ["datasets"], json=self.raw_model, model=Dataset, headers=headers
        )

        self.raw_model = dataset_object.raw_model

        return True

    def delete(self):
        # type: () -> bool
        """Deletes the Dataset.

        Returns:
            bool: True if dataset is deleted.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        return self.connection.api_call("DELETE", ["datasets", self.id], headers=headers)

    def update(self, name=None, description=None, tags=None):
        # type: (str, str, List[str]) -> bool
        """Updates the Dataset.

        Args:
            name (str): Name of the dataset. Defaults to None.
            description (str): Description of the dataset. Defaults to None.
            tags (:obj:`list` of :obj:`str`): List of tags. Defaults to None.

        Returns:
            bool: True, if dataset is updated.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})

        if name is not None:
            self.raw_model["name"] = name
        if description is not None:
            self.raw_model["description"] = description
        if tags is not None:
            self.raw_model["tags"] = tags

        body = self.raw_model

        dataset_object = self.connection.api_call(
            "PUT", ["datasets", self.id], headers=headers, json=body, model=Dataset
        )

        self.raw_model = dataset_object.raw_model

        log.debug("Updated dataset %s with content %s", self.id, dataset_object.__dict__)
        return True

    def refresh(self):
        """Refresh Resource model from API backend.

        Returns:
            bool: True, if it is able to refresh the model,
                False otherwise.
        """
        # type () -> bool
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        dataset_object = self.connection.api_call(
            "GET", ["datasets", self.id], headers=headers, model=Resource
        )

        self.raw_model = dataset_object.raw_model

        return True

    def create_file(self, path, tags=None, description=None):
        # type: (str, List[str], str) -> File
        """Creates File resource in Dataset.

        Args:
            path (str): Path of the file resource.
            tags (:obj:`list` of :obj:`str`): Tags of the file resource.
                Defaults to None.
            description (str): Description of the file resource.
                Defaults to None.

        Returns:
            crux.models.File: File Object.
        """

        headers = Headers({"content-type": "application/json", "accept": "application/json"})

        tags = tags if tags else []

        file_name, folder = split_posixpath_filename_dirpath(path)

        raw_model = {
            "name": file_name,
            "type": "file",
            "tags": tags,
            "description": description,
            "folder": folder,
        }

        file_resource = File(raw_model=raw_model)

        return self.connection.api_call(
            "POST",
            ["datasets", self.id, "resources"],
            json=file_resource.raw_model,
            model=File,
            headers=headers,
        )

    def create_folder(self, path, folder="/", tags=None, description=None):
        # type: (str, str, List[str], str) -> Folder
        """Creates Folder resource in Dataset.

        Args:
            path (str): Path of the Folder resource.
            folder (str): Parent folder of the Folder resource.
                Defaults to /.
            tags (:obj:`list` of :obj:`str`): Tags of the Folder resource.
                Defaults to None.
            description (str): Description of the Folder resource.
                Defaults to None.

        Returns:
            crux.models.Folder: Folder Object.
        """

        headers = Headers({"content-type": "application/json", "accept": "application/json"})

        tags = tags if tags else []

        file_name, folder = split_posixpath_filename_dirpath(path)

        raw_model = {
            "name": file_name,
            "type": "folder",
            "tags": tags,
            "description": description,
            "folder": folder,
        }

        folder_resource = Folder(raw_model=raw_model)

        return self.connection.api_call(
            "POST",
            ["datasets", self.id, "resources"],
            json=folder_resource.raw_model,
            model=Folder,
            headers=headers,
        )

    def _get_resource(self, path, model):
        """Gets the resource object from the string path.

        Args:
            path (str): Resource path.
            crux.models.Resource: Resource model.

        Returns:
            crux.models.Resource: Resource model Object.

        Raises:
            crux.exceptions.CruxResourceNotFoundError: If resource is not found.
        """
        resource_name, folder_path = split_posixpath_filename_dirpath(path)
        resource_gen = self._list_resources(
            folder=folder_path, limit=1, name=resource_name, include_folders=True, model=model,
        )
        try:
            return next(resource_gen)
        except StopIteration as ex:
            log.debug("Parse exception: %s" % str(ex))
            # As of now API can't fetch id from the resource path or name,
            # hence raising the 404 error from the Python client
            raise CruxResourceNotFoundError({"statusCode": 404, "name": resource_name})

    def _resource_exists(self, path):
        # type: (str) -> bool
        """Checks the existence of resource.

        Args:
            path (str): Resource Path.

        Returns:
            bool: True if resource exists.

        Raises:
            crux.exceptions.CruxResourceNotFoundError: If resource is not found.
        """
        try:
            self._get_resource(path=path, model=Resource)
            return True
        except CruxResourceNotFoundError:
            return False

    def get_file(self, path):
        # type: (str) -> File
        """Gets the File resource object.

        Args:
            path (str): File resource path.

        Returns:
            crux.models.File: File Object.
        """
        return self._get_resource(path=path, model=File)

    def get_folder(self, path):
        # type: (str) -> Folder
        """Gets the Folder resource object.

        Args:
            path (str): Folder resource path.

        Returns:
            crux.models.Folder: Folder Object.
        """
        return self._get_resource(path=path, model=Folder)

    def list_resources(
        self, folder="/", cursor=None, limit=1, include_folders=False, sort=None
    ):
        # type: (str, str, int, bool, str) -> Generator[Resource]
        """Lists the resources in Dataset.

        Args:
            folder (str): Folder for which resource should be listed.
                Defaults to /.
            cursor (str): Sets the offset to the page cursor. Defaults to None.
            limit (int): Sets the limit. Defaults to 1.
            include_folders (bool): Sets whether to include folders or not.
                Defaults to False.
            sort (str): Sets whether to sort or not.
                Defaults to None.

        Returns:
            list (:obj:`crux.models.Resource`): List of File resource objects.
        """
        result_gen = self._list_resources(
            sort=sort,
            folder=folder,
            cursor=cursor,
            limit=limit,
            include_folders=include_folders,
            model=Resource,
        )

        for result in result_gen:
            yield result

    def download_files(self, folder, local_path, only_use_crux_domains=None):
        # type: (str, str, bool) -> List[str]
        """Downloads the resources recursively.

        Args:
            folder (str): Crux Dataset Folder from where the
                file resources should be recursively downloaded.
            local_path (str): Local OS Path where the file resources should be downloaded.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Returns:
            list (:obj:`str`): List of location of download files.

        Raises:
            ValueError: If Folder or local_path is None.
            OSError: If local_path is an invalid directory location.
        """
        if folder is None:
            raise ValueError("Folder value shouldn't be empty")

        if local_path is None:
            raise ValueError("Local Path value shouldn't be empty")

        if not os.path.exists(local_path) and not os.path.isdir(local_path):
            raise OSError("local_path is an invalid directory location")

        resources_gen = self._list_resources(
            sort=None,
            folder=folder,
            cursor=None,
            limit=None,
            include_folders=True,
            model=Resource,
        )

        for resource in resources_gen:
            resource_path = posixpath.join(folder, resource.name)
            resource_local_path = os.path.join(local_path, resource.name)
            if resource.type == "folder":
                if not os.path.exists(resource_local_path):
                    os.mkdir(resource_local_path)
                log.debug("Created local directory %s", resource_local_path)
                result_gen = self.download_files(
                    folder=resource_path,
                    local_path=resource_local_path,
                    only_use_crux_domains=only_use_crux_domains,
                )
                for result in result_gen:
                    yield result
            elif resource.type == "file":
                file_resource = File.from_dict(resource.to_dict(), connection=self.connection)
                file_resource.download(
                    resource_local_path, only_use_crux_domains=only_use_crux_domains
                )
                yield resource_local_path
                log.debug("Downloaded file at %s", resource_local_path)

    def upload_files(
        self,
        local_path,
        folder,
        media_type=None,
        description=None,
        tags=None,
        only_use_crux_domains=None,
    ):
        # type: (str, str, str, str, List[str], bool) -> List[File]
        """Uploads the resources recursively.

        Args:
            local_path (str): Local OS Path from where the file resources should be uploaded.
            media_type (str): Content Types of File resources to be uploaded.
                Defaults to None.
            folder (str): Crux Dataset Folder where file resources
                should be recursively uploaded.
            description (str): Description to be set on uploaded resources.
                Defaults to None.
            tags (:obj:`list` of :obj:`str`): Tags to be set on uploaded resources.
                Defaults to None.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Returns:
            list (:obj:`crux.models.File`): List of uploaded file objects.

        Raises:
            ValueError: If folder or local_path is None.
            OSError: If local_path is an invalid directory location.
        """
        tags = tags if tags else []

        uploaded_file_objects = []  # type: List[File]

        if folder is None:
            raise ValueError("Folder value shouldn't be empty")

        if local_path is None:
            raise ValueError("Local Path value shouldn't be empty")

        if not os.path.exists(local_path) and not os.path.isdir(local_path):
            raise OSError("local_path is an invalid directory location")

        for content in os.listdir(local_path):
            content_local_path = os.path.join(local_path, content)
            content_path = posixpath.join(folder, content)
            if os.path.isdir(content_local_path):
                self.create_folder(path=content_path, tags=tags, description=description)
                log.debug("Created folder %s in dataset %s", content_path, self.id)
                uploaded_file_objects += self.upload_files(
                    media_type=media_type,
                    folder=content_path,
                    local_path=content_local_path,
                    tags=tags,
                    description=description,
                    only_use_crux_domains=only_use_crux_domains,
                )

            elif os.path.isfile(content_local_path):
                fil_o = self.upload_file(
                    content_local_path,
                    content_path,
                    media_type=media_type,
                    tags=tags,
                    description=description,
                    only_use_crux_domains=only_use_crux_domains,
                )
                uploaded_file_objects.append(fil_o)
                log.debug("Uploaded file %s in dataset %s", content_path, self.id)

        return uploaded_file_objects

    def list_files(self, sort=None, folder="/", cursor=None, limit=100):
        # type: (str, str, str, int) -> List[File]
        """Lists the files.

        Args:
            sort (str): Sets whether to sort or not. Defaults to None.
            folder (str): Folder for which resource should be listed.
                Defaults to /.
            cursor (str): Sets the offset to the page cursor. Defaults to None.
            limit (int): Sets the limit. Defaults to 100.

        Returns:
            list (:obj:`crux.models.File`): List of File objects.
        """
        resource_list_gen = self._list_resources(
            sort=sort,
            folder=folder,
            cursor=cursor,
            limit=limit,
            include_folders=False,
            model=File,
        )

        for resource in resource_list_gen:
            if resource.type == "file":
                yield resource

    def _list_resources(
        self,
        folder="/",
        cursor=None,
        limit=1,
        include_folders=False,
        name=None,
        model=None,
        sort=None,
    ):

        headers = Headers({"content-type": "application/json", "accept": "application/json"})

        params = {"datasetId": self.id, "folder": folder}

        if cursor:
            params["cursor"] = cursor

        if sort:
            params["sort"] = sort

        if name:
            params["name"] = name

        if include_folders:
            params["includeFolders"] = "true"
        else:
            params["includeFolders"] = "false"

        retrieved = 0
        pagesize = 500
        paginate = {}
        while limit is None or retrieved < limit:
            params["limit"] = None if limit is None else min(pagesize, limit - retrieved)

            resp = self.connection.api_call(
                "GET",
                ["resources"],
                params=params,
                model=model,
                headers=headers,
                paginate=paginate,
            )
            resp_count = len(resp)

            if resp_count == 0:
                break

            for resource in resp:
                yield resource

            retrieved += resp_count
            params["cursor"] = paginate["cursor"]

    def upload_file(
        self,
        src,
        dest,
        media_type=None,
        description=None,
        tags=None,
        only_use_crux_domains=None,
    ):
        # type: (Union[IO, str], str, str, str, List[str], bool) -> File
        """Uploads the File.

        Args:
            src (str or file): Local OS path whose content
                is to be uploaded to file resource.
            dest (str): File resource path.
            media_type (str): Content type of the file. Defaults to None.
            description (str): Description of the file. Defaults to None.
            tags (:obj:`list` of :obj:`str`): Tags to be attached to the file resource.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Returns:
            crux.models.File: File Object.
        """
        tags = tags if tags else []

        file_resource = self.create_file(tags=tags, description=description, path=dest)

        try:
            return file_resource.upload(
                src, media_type=media_type, only_use_crux_domains=only_use_crux_domains
            )
        except (CruxClientError, CruxAPIError, IOError):
            file_resource.delete()
            raise

    def add_permission_to_resources(
        self,
        identity_id,
        permission,
        resource_paths=None,
        resource_objects=None,
        resource_ids=None,
    ):
        # type: (str, str, List[str], List[Union[File,Folder]], List[str]) -> bool
        """Adds permission to all or specific Dataset resources.

        Args:
            identity_id (str): Identity Id to be set.
            permission (str): Permission to be set.
            resource_paths (:obj:`list` of :obj:`str`): List of resource paths on which the
                permission should be applied. If none of resource_paths,
                resource_objects or resource_ids parameter is set,
                then it will apply the permission to whole dataset.
            resource_objects (:obj:`list` of :obj:`crux.models.Resource`): List of
                resource objects on which the permission should be applied.
                Overrides resource_paths.
                If none of resource_paths, resource_objects or
                resource_ids parameter is set, then it will apply the
                permission to whole dataset.
            resource_ids (:obj:`list` of :obj:`str`): List of resource ids on which
                permission should be applied. Overrides resource_pathss and resource_objects.
                If none of resource_paths,
                resource_objects or resource_ids parameter is set,
                then it will apply the permission to whole dataset.

        Returns:
            bool: True if permission is applied.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        body = {
            "identityId": identity_id,
            "permission": permission,
            "action": "add",
        }  # type: Dict[str, Union[str, List[str]]]

        if resource_ids or resource_objects or resource_paths:
            if resource_paths:
                log.debug(
                    "Add permission %s to %s for resource paths", permission, identity_id,
                )
                resource_ids = list()
                for resource_path in resource_paths:
                    resource_object = self._get_resource(path=resource_path, model=Resource)
                    resource_ids.append(resource_object.id)
                body["resourceIds"] = resource_ids

            if resource_objects:
                log.debug(
                    "Add permissions %s to %s for resource objects", permission, identity_id,
                )
                resource_ids = list()
                for resource_object in resource_objects:
                    resource_ids.append(resource_object.id)
                body["resourceIds"] = resource_ids

            if resource_ids:
                log.debug("Add permissions %s to %s for resource ids", permission, identity_id)
                body["resourceIds"] = resource_ids
        else:
            log.debug(
                "Add permission %s to %s for dataset %s", permission, identity_id, self.id,
            )
            body["datasetId"] = self.id

        return self.connection.api_call(
            "POST", ["permissions", "bulk"], headers=headers, json=body
        )

    def delete_permission_from_resources(
        self,
        identity_id,
        permission,
        resource_paths=None,
        resource_objects=None,
        resource_ids=None,
    ):
        # type: (str, str, List[str], List[Union[File,Folder]], List[str]) -> bool
        """Method which deletes permission from all or specific Dataset resources.

        Args:
            identity_id (str): Identity Id for the deletion.
            permission (str): Permission for the deletion.
            resource_paths (:obj:`list` of :obj:`crux.models.Resource`): List of
                resource path from which the permission should be deleted.
                If none of resource_paths,
                resource_objects or resource_ids parameter is set,
                then it will delete the permission from whole dataset.
            resource_objects (:obj:`list` of :obj:`crux.models.Resource`): List of
                resource objects from which the permission should be deleted.
                Overrides resource_paths.
                If none of resource_paths, resource_objects or resource_ids parameter is
                set, then it will delete the permission from whole dataset.
            resource_ids (:obj:`list` of :obj:`crux.models.Resource`): List of
                resource ids from which the permission should be deleted.
                Overrides resource_paths and resource_objects.
                If none of resource_paths, resource_objects or
                resource_ids parameter is set, then it will delete the
                permission from whole dataset.

        Returns:
            bool: True if it is able to delete the permission.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        body = {
            "identityId": identity_id,
            "permission": permission,
            "action": "delete",
        }  # type: Dict[str, Union[str, List[str]]]

        if resource_ids or resource_objects or resource_paths:
            if resource_paths:
                log.debug(
                    "Delete permission %s for %s from resource paths", permission, identity_id,
                )
                resource_ids = list()
                for resource_path in resource_paths:
                    resource_object = self._get_resource(path=resource_path, model=Resource)
                    resource_ids.append(resource_object.id)
                body["resourceIds"] = resource_ids

            if resource_objects:
                log.debug(
                    "Delete permission %s for %s from resource objects",
                    permission,
                    identity_id,
                )
                resource_ids = list()
                for resource_object in resource_objects:
                    resource_ids.append(resource_object.id)
                body["resourceIds"] = resource_ids

            if resource_ids:
                log.debug(
                    "Delete permission %s for %s from resource ids", permission, identity_id,
                )
                body["resourceIds"] = resource_ids
        else:
            log.debug(
                "Delete permission %s for %s from dataset %s",
                permission,
                identity_id,
                self.id,
            )
            body["datasetId"] = self.id

        return self.connection.api_call(
            "POST", ["permissions", "bulk"], headers=headers, json=body
        )

    def add_permission(self, identity_id, permission):
        # type: (str, str) -> Union[bool, Permission]
        """Adds permission to the Dataset.

        Args:
            identity_id: Identity Id to be set.
            permission: Permission to be set.

        Returns:
            crux.models.Permission: Permission Object.
        """
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }  # type: MutableMapping[Text, Text]
        return self.connection.api_call(
            "PUT",
            ["permissions", self.id, identity_id, permission],
            model=Permission,
            headers=headers,
        )

    def delete_permission(self, identity_id, permission):
        # type: (str, str) -> bool
        """Deletes permission from the Dataset.

        Args:
            identity_id (str): Identity Id for the deletion.
            permission (str): Permission for the deletion.

        Returns:
            bool: True if it is able to delete it.
        """
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }  # type: MutableMapping[Text, Text]
        return self.connection.api_call(
            "DELETE", ["permissions", self.id, identity_id, permission], headers=headers
        )

    def list_permissions(self):
        # type: () -> List[Permission]
        """Lists the permission on the Dataset.

        Returns:
            list (:obj:`crux.models.Permission`): List of Permission Objects.
        """
        headers = {"Accept": "application/json"}  # type: MutableMapping[Text, Text]
        return self.connection.api_call(
            "GET", ["datasets", self.id, "permissions"], model=Permission, headers=headers,
        )

    def add_label(self, label_key, label_value):
        # type: (str, str) -> bool
        """Adds label to Dataset.

        Args:
            label_key (str): Label Key for Dataset.
            label_value (str): Label Value for Dataset.

        Returns:
            bool: True if labels are added.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        return self.connection.api_call(
            "PUT", ["datasets", self.id, "labels", label_key, label_value], headers=headers,
        )

    def delete_label(self, label_key):
        # type: (str) -> bool
        """Deletes label from Dataset.

        Args:
            label_key (str): Label Key for Dataset.

        Returns:
            bool: True if labels are deleted.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        return self.connection.api_call(
            "DELETE", ["datasets", self.id, "labels", label_key], headers=headers
        )

    def get_label(self, label_key):
        # type: (str) -> Label
        """Gets label value of Dataset.

        Args:
            label_key (str): Label Key for Dataset.

        Returns:
            crux.models.Label: Label Object.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        return self.connection.api_call(
            "GET", ["datasets", self.id, "labels", label_key], headers=headers, model=Label,
        )

    def find_resources_by_label(self, predicates, max_per_page=1000):
        # type: (List[Dict[str,str]],int)->Iterator[Union[File,Folder]]
        """Method which searches the resouces for given labels in Dataset

        Each predicate can be either:

        - Lexicographical equal
        - Lexicographical less than
        - Lexicographical less than or equal to
        - Lexicographical greater than
        - Lexicographical greater than or equal to
        - A list of OR predicates
        - A list of AND predicates

        .. code-block:: python

            predicates = [
                {"op": "eq", "key": "key1", "val": "abcd"},
                {"op": "ne", "key": "key1", "val": "zzzz"},
                {"op": "lt", "key": "key1", "val": "abd"},
                {"op": "gt", "key": "key1", "val": "abc"},
                {"op": "lte", "key": "key1", "val": "abd"},
                {"op": "gte", "key": "key1", "val": "abc"},
                {"op": "or", "in":
                    [
                        {"op": "eq", "key": "key1", "val": "abcd"},
                        # more OR predicates...
                    ]
                },
                {"op": "and", "in":
                    [
                        {"op": "eq", "key": "key1", "val": "abcd"},
                        # more AND predicates...
                    ]
                }
            ]

        Args:
            predicates (:obj:`list` of :obj:`dict`): List of dictionary predicates
                for finding resources.
            max_per_page (int): Pagination limit. Defaults to 1000.

        Returns:
            list (:obj:`crux.models.Resource`): List of resource matching the query parameters.

        Example:
            .. code-block:: python

                from crux import Crux

                conn = Crux()
                dataset_object = conn.get_dataset(id="dataset_id")
                predicates=[
                    {"op":"eq","key":"test_label1","val":"test_value1"}
                ]
                resource_objects = dataset_object.find_resources_by_label(
                    predicates=predicates
                )
        """

        predicates = predicates if predicates else []

        query_params = {"limit": max_per_page}

        predicates_query = {"basic_query": predicates}  # type: Dict[str, List[Dict[str,str]]]

        after = None
        headers = Headers({"content-type": "application/json", "accept": "application/json"})

        while True:

            if after:
                query_params["after"] = after

            response = self.connection.api_call(
                "POST",
                ["datasets", self.id, "labels", "search"],
                headers=headers,
                json=predicates_query,
                params=query_params,
            )

            resource_list = response.json().get("results")
            if resource_list:
                after = resource_list[-1].get("resourceId")
                for resource in resource_list:
                    obj = get_resource_object(
                        resource_type=resource.get("type"),
                        data=resource,
                        connection=self.connection,
                    )
                    yield obj
            else:
                return

    def stitch(
        self, source_resources, destination_resource, labels=None, tags=None, description=None,
    ):
        # type: (List[Union[str, File]], str, str, List[str], str) -> Tuple[File, str]
        """Method which stitches multiple Avro resources into single Avro resource

        Args:
            source_resources (:obj:`list` of :obj:`str` or :obj:`file`): List of resource paths
                which are to be stitched.
            destination_resource (str): Resource Path to load the stitched output
            labels (dict): Key/Value labels that should be applied to stitched resource
            tags (:obj:`list` of :obj:`str`): List of tags to be applied
                on destination resource.
                Taken into consideration if resource is required to be created.
            description (str): Description to be applied created destination.
                Taken into consideration if resource is required to be created.

        Returns:
            tuple (:obj:`crux.models.File`, :obj:`str`): File object of destination resource.
                    Job ID for background running job.
        Raises:
            TypeError: Source and Destination resource should be of type File or String
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        source_resource_ids = list()
        for resource in source_resources:
            if isinstance(resource, File):
                log.debug("Stitch source resources are of type crux.models.File")
                source_resource_ids.append(resource.id)
            elif isinstance(resource, (str, unicode)):
                log.debug("Stitch source resources are of type string")
                resource_object = self._get_resource(path=resource, model=File)
                source_resource_ids.append(resource_object.id)
            else:
                raise TypeError(
                    "Invalid Type. It should be File resource object or path string"
                )

        if isinstance(destination_resource, File):
            log.debug("Stitch destination resource is of type crux.models.File")
            destination_file_object = destination_resource
        elif isinstance(destination_resource, str):
            log.debug("Stitch destination resource is of type string")
            if self._resource_exists(path=destination_resource):
                destination_file_object = self._get_resource(
                    path=destination_resource, model=File
                )
            else:
                log.debug("Creating file resource at %s", destination_resource)
                destination_file_object = self.create_file(
                    path=destination_resource,
                    description=description,
                    tags=tags if tags else [],
                )
        else:
            raise TypeError("Invalid Type. It should be File resource object or path string")

        data = {
            "sourceResourceIds": source_resource_ids,
            "destinationResourceId": destination_file_object.id,
            "labelsToApply": labels if labels else {},
        }
        response = self.connection.api_call(
            "POST", ["datasets", self.id, "stitch"], headers=headers, json=data
        )

        raw_json = response.json().get("destinationResource")

        file_object = Resource.from_dict(raw_json, connection=self.connection)

        job_id = response.json().get("jobId")

        return (file_object, job_id)

    def get_stitch_job(self, job_id):
        # type: (str) -> StitchJob
        """Stitch Job Details.

        Args:
            job_id (str): Job ID of the Stitch Job.

        Returns:
            crux.models.StitchJob: StitchJob object.
        """
        headers = Headers({"content-type": "application/json", "accept": "application/json"})
        return self.connection.api_call(
            "GET", ["datasets", "stitch", job_id], headers=headers, model=StitchJob
        )

    def get_delivery(self, delivery_id):
        # type: (str) -> Delivery
        """Gets Delivery object.

        Args:
            delivery_id (str): Delivery ID.

        Returns:
            crux.models.Delivery: Delivery Object.
        Raises:
            ValueError: If delivery_id value is invalid.
        """
        headers = Headers({"accept": "application/json"})

        if not DELIVERY_ID_REGEX.match(delivery_id):
            raise ValueError("Value of delivery_id is invalid")

        return self.connection.api_call(
            "GET", ["deliveries", self.id, delivery_id], headers=headers, model=Delivery
        )

    def get_ingestions(
        self, start_date=None, end_date=None, delivery_status=None, use_cache=None
    ):
        # type: (str, str, str, bool) -> Iterator[Ingestion]
        """Gets Ingestions.

        Args:
            start_date (str): ISO format start time.
            end_date (str): ISO format end time.
            delivery_status (str): Delivery status enum
            use_cache (bool): Preference to set cached response

        Returns:
            crux.models.Delivery: Delivery Object.
        """
        headers = Headers({"accept": "application/json"})

        params = {}
        params["start_date"] = start_date
        params["end_date"] = end_date

        if delivery_status:
            params["delivery_status"] = delivery_status.upper()

        if use_cache is not None:
            params["useCache"] = use_cache

        response = self.connection.api_call(
            "GET", ["deliveries", self.id, "ids"], headers=headers, params=params
        )

        response_json = response.json()
        if isinstance(response_json, dict):
            all_deliveries = response_json.get("delivery_ids")
        else:
            all_deliveries = response_json

        ingestion_set = defaultdict(set)  # type: DefaultDict[str, Set]

        for delivery in all_deliveries:
            if not DELIVERY_ID_REGEX.match(delivery):
                raise ValueError("Value of delivery_id is invalid")
            ingestion_id, version_id = delivery.split(".")
            ingestion_set[ingestion_id].add(int(version_id))

        for ingestion_id in ingestion_set:
            obj = Ingestion.from_dict(
                {
                    "ingestionId": ingestion_id,
                    "versions": ingestion_set[ingestion_id],
                    "datasetId": self.id,
                }
            )
            obj.connection = self.connection
            yield obj

    def get_latest_files(
        self,
        frames=None,
        file_format=MediaType.AVRO.value,
        delivery_status=None,
        use_cache=None,
    ):

        # type: (Optional[Union[str, List]], str, str, bool) -> Iterator[File]
        """Get the latest dataset file resources. The latest supplier_implied_dt with the
        best single delivery version is selected.

        Args:
            frames (str, list): filter for selected frames
            file_format (str): File format of delivery.
            delivery_status (str): Delivery status enum
            use_cache (bool): Preference to set cached response

        Returns:
            list (:obj:`crux.models.File`): List of file resources.
        """

        # look back a couple extra days in case query is performed over the weekend
        got_files = False
        for lookback in [1 + 2, 14 + 2, 31 + 2, 180 + 2, 366 + 2]:
            start_date = datetime.utcnow() - timedelta(days=lookback)

            series = self.get_files_range(
                start_date=start_date.isoformat(),
                frames=frames,
                file_format=file_format,
                latest_only=True,
                delivery_status=delivery_status,
                use_cache=use_cache,
            )
            for item in series:
                got_files = True
                yield item

            if got_files:
                break

    def get_files_range(
        self,
        start_date,  # type: Union[datetime, str]
        end_date=None,  # type: Optional[Union[datetime, str]]
        frames=None,  # type: Optional[Union[str, list]]
        file_format=MediaType.AVRO.value,  # type: str
        dayfirst=False,  # type: bool
        yearfirst=False,  # type: bool
        latest_only=False,  # type: bool
        delivery_status=None,  # type: str
        use_cache=None,  # type: bool
    ):
        # type: (...) -> Iterator[File]
        """Get a set of dataset file resources. The best single delivery version for each
        supplier_implied_dt is computed for the selected time range.

        Args:
            start_date (str): ISO format start datetime or any paresable date string.
            end_date (str): ISO format end datetime or any parseable date string.
            delivery_status (str): Delivery status enum
            frames (str, list): filter for selected frames
            file_format (str): File format of delivery.

        Returns:
            list (:obj:`crux.models.File`): List of file resources.
        """

        if isinstance(frames, list):
            frames = [x.upper() for x in frames]
        elif isinstance(frames, str):
            frames = [frames.upper()]
        elif frames is not None:
            raise ValueError("Value of frames is invalid")

        if isinstance(file_format, MediaType):
            file_format = file_format.value
        else:
            if file_format not in [item.value for item in MediaType]:
                raise ValueError("Value of file_format is invalid")

        if isinstance(start_date, datetime):
            stdt = start_date
        elif isinstance(start_date, str):
            try:
                stdt = parser.parse(start_date, dayfirst=dayfirst, yearfirst=yearfirst)
            except:
                raise ValueError("Value of start_date is invalid")
        else:
            raise ValueError("start_date must be str or datetime")

        if end_date is None:
            enddt = None
        elif isinstance(end_date, datetime):
            enddt = end_date
        elif isinstance(end_date, str):
            try:
                enddt = parser.parse(end_date, dayfirst=dayfirst, yearfirst=yearfirst)
            except:
                raise ValueError("Value of end_date is invalid")
        else:
            raise ValueError("date must be str or datetime")

        headers = Headers({"accept": "application/json"})

        delivery_status = "DELIVERY_SUCCEEDED" if not delivery_status else delivery_status
        params = {"start_date": stdt, "end_date": enddt, "delivery_status": delivery_status}

        if use_cache is not None:
            params["useCache"] = use_cache

        response = self.connection.api_call(
            "GET", ["deliveries", self.id, "ids"], headers=headers, params=params
        )
        delivery_set = {}
        deliveryid_mapping = {}

        response_json = response.json()
        if isinstance(response_json, dict):
            all_deliveries = response_json.get("delivery_ids")
        else:
            all_deliveries = response_json

        for delivery_id in all_deliveries:
            delivery_resources = []
            if not DELIVERY_ID_REGEX.match(delivery_id):
                raise ValueError("Value of delivery_id is invalid")
            params = {"delivery_resource_format": file_format}
            response = self.connection.api_call(
                "GET", ["deliveries", self.id, delivery_id, "data"], params=params
            )
            data = response.json()
            if data["latest_health_status"] != "DELIVERY_SUCCEEDED" or not data["resources"]:
                continue
            for item in data["resources"]:
                if frames is not None and item["frame_id"].upper() not in frames:
                    continue
                delivery_resources.append(item)
                deliveryid_mapping[item["resource_id"]] = delivery_id
            delivery_set[delivery_id] = {
                "deliver_id": delivery_id,
                "resource_ids": delivery_resources,
                "files": [],
                "supplier_implied_dt": None,
                "ingestion_time": None,
            }

        if delivery_set and not sum([x["resource_ids"] for x in delivery_set.values()], []):
            raise ValueError("No file resources found for selected frames")

        for item in self.get_resources_batch(list(deliveryid_mapping.keys())):
            delivery_id = deliveryid_mapping[item.id]
            if delivery_id != ".".join(
                [item.labels["ingestion_id"], item.labels["version_id"]]
            ):
                log.debug(
                    "File resource has incorrect labels for delivery_id=%s %s %s",
                    delivery_id,
                    item.labels["ingestion_id"],
                    item.labels["version_id"],
                )
                item.labels["ingestion_id"] = delivery_id.split(".")[0]
                item.labels["version_id"] = delivery_id.split(".")[1]

            delivery_set[delivery_id]["files"].append(item)
            if delivery_set[delivery_id]["supplier_implied_dt"] is None:
                delivery_set[delivery_id]["supplier_implied_dt"] = item.supplier_implied_dt
                delivery_set[delivery_id]["ingestion_time"] = item.ingestion_time

        best_deliveries = {}
        for item in delivery_set.values():
            if item["supplier_implied_dt"] is None:
                continue
            dt = item["supplier_implied_dt"]
            if (
                dt not in best_deliveries
                or item["ingestion_time"] > best_deliveries[dt]["ingestion_time"]
                or (
                    item["ingestion_time"] == best_deliveries[dt]["ingestion_time"]
                    and item["delivery_id"] > best_deliveries[dt]["delivery_id"]
                )
            ):
                best_deliveries[dt] = item

        if latest_only:
            if best_deliveries.keys():
                dt = sorted(best_deliveries.keys())[-1]
                for file in best_deliveries[dt]["files"]:
                    yield file
        else:
            for dt in sorted(best_deliveries.keys()):
                for file in best_deliveries[dt]["files"]:
                    yield file

    def get_resources_batch(self, resource_ids):
        # type: (list) -> Iterator[File]
        """Gets resource metadata.

        Args:
            resource_ids (list): List of resource IDs

        Returns:
            list (:obj:`crux.models.File`): List of file resources.
        """
        headers = Headers({"accept": "application/json"})
        limit = 250
        for i in range(0, len(resource_ids), limit):
            chunk = resource_ids[i : i + limit]
            response = self.connection.api_call(
                "POST",
                ["resources", "get-batch"],
                params={"limit": limit + 1},
                headers=headers,
                data=json.dumps(chunk),
            )
            for resource in response.json():
                obj = File(raw_model=resource)
                obj.connection = self.connection
                yield obj



"""Module contains File model."""

from typing import Any, Dict, IO, Iterable, List, Union  # noqa: F401

from google.resumable_media.common import (  # type: ignore
    DataCorruption,
    InvalidResponse,
)
from google.resumable_media.requests import (  # type: ignore
    ChunkedDownload,
    ResumableUpload,
)
from requests.exceptions import (
    ConnectTimeout,
    HTTPError,
    ProxyError,
    ReadTimeout,
    SSLError,
    TooManyRedirects,
)

from crux._compat import unicode
from crux._utils import (
    create_logger,
    DEFAULT_CHUNK_SIZE,
    get_session,
    Headers,
    ResumableUploadSignedSession,
    valid_chunk_size,
)
from crux.exceptions import (
    CruxClientConnectionError,
    CruxClientError,
    CruxClientHTTPError,
    CruxClientTimeout,
    CruxClientTooManyRedirects,
)
from crux.models.resource import MediaType, Resource


log = create_logger(__name__)


class File(Resource):
    """File Model."""

    def _get_signed_url(self):
        headers = Headers(
            {"content-type": "application/json", "accept": "application/json"}
        )
        response = self.connection.api_call(
            "POST", ["resources", self.id, "content-url"], headers=headers, json={}
        )

        url = response.json().get("url")

        if not url:
            raise KeyError(
                "Signed URL missing in response for resource {id}".format(id=self.id)
            )

        return url

    def _dl_signed_url(self, file_obj, chunk_size=DEFAULT_CHUNK_SIZE):
        """Download from signed URL using requests directly, not google-resumable-media."""
        signed_url = self._get_signed_url()

        log.trace("Using direct signed url: %s", signed_url)

        transport = get_session(proxies=self.connection.crux_config.proxies)

        log.debug("Using Proxies %s for downloading", transport.proxies)

        try:
            with transport as session:
                response = session.get(signed_url, stream=True)
                response.raise_for_status()
                for chunk in response.iter_content(chunk_size=chunk_size):
                    file_obj.write(chunk)
        except HTTPError as err:
            raise CruxClientHTTPError(str(err), err.response)
        except TooManyRedirects as err:
            raise CruxClientTooManyRedirects(str(err))
        except (ProxyError, SSLError) as err:
            raise CruxClientConnectionError(str(err))
        except (ConnectTimeout, ReadTimeout) as err:
            raise CruxClientTimeout(str(err))

        return True

    def _dl_signed_url_resumable(self, file_obj, chunk_size=DEFAULT_CHUNK_SIZE):
        """Download from signed URL using google-resumable-media."""
        signed_url = self._get_signed_url()

        log.trace("Using resumable signed url: %s", signed_url)

        transport = get_session(proxies=self.connection.crux_config.proxies)

        log.debug("Using Proxies %s for downloading", transport.proxies)

        # Track how many bytes the client has downloaded since the last time they
        # got a new signed URL, and how many times that got a new URL without
        # downloading more bytes.
        # This is to prevent infinite loops without progress.
        bytes_at_last_refresh = 0
        refreshes_without_progress = 0
        # Track how much has been downloaded from each signed URL
        total_bytes_from_urls = [0]
        fetched_signed_urls = 0
        max_url_refreshes_without_progress = 5
        max_url_refreshes = 100

        download = ChunkedDownload(signed_url, chunk_size, file_obj)

        log.debug("Starting download using signed url for resource %s", self.id)

        while not download.finished:
            try:
                # This downloads a chunk and writes it to file_object
                download.consume_next_chunk(transport)
                total_bytes_from_urls[-1] = download.bytes_downloaded
            # Catch the signed URL expiring
            except InvalidResponse:
                # Limit total new URL(s)
                if fetched_signed_urls >= max_url_refreshes:
                    raise CruxClientError("Exceeded max new Signed URLs")
                sum_total_bytes_from_urls = sum(total_bytes_from_urls)
                # Check if download has made progress downloading since last time we got URL
                if not sum_total_bytes_from_urls > bytes_at_last_refresh:
                    # Limit new URLs without making progress downloading
                    if refreshes_without_progress <= max_url_refreshes_without_progress:
                        new_signed_url = self._get_signed_url()
                        fetched_signed_urls += 1
                        log.debug(
                            "fetched_signed_urls count for download is %s",
                            fetched_signed_urls,
                        )
                        total_bytes_from_urls.append(0)
                        refreshes_without_progress += 1
                        log.debug(
                            "refreshes_without_progress count for download is %s",
                            refreshes_without_progress,
                        )
                        bytes_at_last_refresh = sum_total_bytes_from_urls
                    else:
                        # Exceeded max new signed URLs without progress
                        raise CruxClientError(
                            "Exceeded max new Signed URLs without progress"
                        )
                else:
                    refreshes_without_progress = 0
                    log.debug("Fetching new singed url")
                    new_signed_url = self._get_signed_url()
                    log.trace("New signed url: %s", new_signed_url)
                    fetched_signed_urls += 1
                    log.debug(
                        "fetched_signed_urls count for download is %s",
                        fetched_signed_urls,
                    )
                    total_bytes_from_urls.append(0)
                    bytes_at_last_refresh = sum_total_bytes_from_urls

                # Replace the download object with a new one, using a new signed URL,
                # but start where the last download object left off.
                log.debug(
                    "Resuming download with new_signed_url %s starting at %s bytes",
                    new_signed_url,
                    sum_total_bytes_from_urls,
                )
                download = ChunkedDownload(
                    new_signed_url,
                    chunk_size,
                    file_obj,
                    start=sum_total_bytes_from_urls,
                )
            except DataCorruption as err:
                raise CruxClientError(err)

        # Closing the session as it is not closed by Resumable Media Lib.
        transport.close()
        log.debug("Download completed using signed url for resource %s", self.id)

        return True

    def iter_content(self, chunk_size=DEFAULT_CHUNK_SIZE, only_use_crux_domains=None):
        # type: (int, bool) -> Iterable[str]
        """Streams the file resource.

        Args:
            chunk_size (int): Chunk Size for the stream.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Yields:
            bytes: Bytes of file resource.

        Raises:
            ValueError: If chunk_size is not multiple of 256 KiB.
        """

        headers = Headers({"accept": "*/*"})

        if not valid_chunk_size(chunk_size):
            raise ValueError("chunk_size should be multiple of 256 KiB")

        # If we must use only Crux domains, download via the API.
        if only_use_crux_domains is None:
            only_use_crux_domains = self.connection.crux_config.only_use_crux_domains

        if only_use_crux_domains:

            log.debug("Using Crux Domain for streaming file resource %s", self.id)
            data = self.connection.api_call(
                "GET", ["resources", self.id, "content"], headers=headers, stream=True
            )
            # Performing early return to avoid below computation
            return data.iter_content(chunk_size=chunk_size)

        log.debug("Using Resumable Signed url for streaming file resource %s", self.id)

        signed_url = self._get_signed_url()
        session = get_session(proxies=self.connection.crux_config.proxies)
        data = session.get(signed_url, stream=True)

        return data.iter_content(chunk_size=chunk_size)

    def _download_file(
        self, file_obj, chunk_size=DEFAULT_CHUNK_SIZE, only_use_crux_domains=None
    ):

        # If size is None it means the file has been created,
        # but no contents uploaded. This is similar to doing touch on command line,
        # it creates an empty file.
        # By returning True here we effectively "download" and write an empty file,
        # which matches file system mechanics of creating a file without writing contents
        if self.size is None:
            log.debug("File resource %s is of size None", self.id)
            return True

        # google-resumable-media has a bug where is expects the 'content-range' even
        # for 200 OK responses, which happens when the range is larger than the size.
        # There isn't much point in using resumable media for small files.
        # Make sure size is greater than 2x chunk_size if Google Resumable media is
        # to be used.
        small_enough = self.size < (chunk_size * 2)

        # If we must use only Crux domains, download via the API.
        if only_use_crux_domains is None:
            only_use_crux_domains = self.connection.crux_config.only_use_crux_domains

        if only_use_crux_domains:
            log.debug("Using Crux Domain for downloading file resource %s", self.id)
            return self._download(
                file_obj=file_obj, media_type=None, chunk_size=chunk_size
            )
        # Use requests directly for small files.
        elif small_enough:
            log.debug(
                "Using Direct Signed url for downloading file resource %s", self.id
            )
            return self._dl_signed_url(file_obj=file_obj, chunk_size=chunk_size)
        # Use google-resumable-media for large files
        else:
            log.debug(
                "Using Resumable Signed url for downloading file resource %s", self.id
            )
            return self._dl_signed_url_resumable(
                file_obj=file_obj, chunk_size=chunk_size
            )

    def download(self, dest, chunk_size=DEFAULT_CHUNK_SIZE, only_use_crux_domains=None):
        # type: (str, int, bool) -> bool
        """Downloads the file resource.

        Args:
            dest (str or file): Local OS path at which file resource will be downloaded.
            chunk_size (int): Number of bytes to be read in memory.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Returns:
            bool: True if it is downloaded.

        Raises:
            TypeError: If dest is not a file like or string type.
        """
        if not valid_chunk_size(chunk_size):
            raise ValueError("chunk_size should be multiple of 256 KiB")

        if hasattr(dest, "write"):
            return self._download_file(
                dest, chunk_size=chunk_size, only_use_crux_domains=only_use_crux_domains
            )
        elif isinstance(dest, (str, unicode)):
            with open(dest, "wb") as file_obj:
                return self._download_file(
                    file_obj,
                    chunk_size=chunk_size,
                    only_use_crux_domains=only_use_crux_domains,
                )
        else:
            raise TypeError("Invalid Data Type for dest: {}".format(type(dest)))

    def _ul_signed_url_resumable(self, file_obj, media_type):

        headers = Headers(
            {
                "content-type": "application/json",
                "accept": "application/json",
                "x-upload-content-type": media_type,
            }
        )

        upload_session_response = self.connection.api_call(
            "POST",
            ["resources", self.id, "upload-session-start"],
            headers=headers,
            json={},
        )
        log.debug("Fetched upload session url for resource %s", self.id)

        upload_response_json = upload_session_response.json()

        signed_url = upload_response_json.get("signedURL").get("url")
        if not signed_url:
            raise KeyError(
                "Signed URL missing in response for resource {id}".format(id=self.id)
            )

        log.trace("Using Resumable upload signed url: %s", signed_url)

        signed_url_headers = Headers(
            upload_response_json.get("signedURL").get("headers")
        )

        log.trace("Signed url headers: %s", signed_url_headers)

        if not signed_url_headers:
            raise KeyError(
                "Signed URL Headers missing in response for resource {id}".format(
                    id=self.id
                )
            )

        session_id = upload_response_json.get("sessionId")

        if not session_id:
            raise KeyError(
                "sessionId Header missing in response for resource {id}".format(
                    id=self.id
                )
            )

        upload = ResumableUpload(signed_url, DEFAULT_CHUNK_SIZE)

        metadata = {"name": self.name}

        transport = get_session(
            session_class=ResumableUploadSignedSession,
            proxies=self.connection.crux_config.proxies,
        )

        transport.headers = signed_url_headers

        log.debug("Using Proxies %s for uploading", transport.proxies)

        log.debug("Initiating upload for resource %s", self.id)

        upload.initiate(
            transport, file_obj, metadata, signed_url_headers["content-type"]
        )

        log.debug("Starting upload using signed url for resource %s", self.id)

        try:
            while not upload.finished:
                if upload.invalid:
                    upload.recover(transport)
                upload.transmit_next_chunk(transport)
        except InvalidResponse as err:
            raise CruxClientError(err)

        log.debug("Upload completed using signed url for resource %s", self.id)

        payload = {"sessionId": session_id}
        return self.connection.api_call(
            "POST",
            ["resources", self.id, "upload-session-complete"],
            headers=headers,
            json=payload,
        )

    def _upload(self, file_obj, media_type, only_use_crux_domains=None):

        if only_use_crux_domains is None:
            only_use_crux_domains = self.connection.crux_config.only_use_crux_domains

        if only_use_crux_domains:
            log.debug("Using Crux Domain for uploading file resource %s", self.id)
            headers = Headers(
                {"content-type": media_type, "accept": "application/json"}
            )
            return self.connection.api_call(
                "PUT",
                ["resources", self.id, "content"],
                data=file_obj,
                headers=headers,
                model=File,
            )

        else:
            log.debug("Using Signed url for uploading file resource %s", self.id)
            return self._ul_signed_url_resumable(file_obj, media_type)

    def upload(self, src, media_type=None, only_use_crux_domains=None):
        # type: (Union[IO, str], str, bool) -> File
        """Uploads the content to empty file resource.

        Args:
            src (str or file): Local OS path whose content is to be uploaded.
            media_type (str): Content type of the file. Defaults to None.
            only_use_crux_domains (bool): True if content is required to be downloaded
                from Crux domains else False.

        Returns
            File: File model object.

        Raises:
            TypeError: If src type is invalid.
        """

        if hasattr(src, "read"):

            if media_type is None:
                media_type = MediaType.detect(getattr(src, "name"))

            upload_result = self._upload(
                src, media_type=media_type, only_use_crux_domains=only_use_crux_domains
            )

        elif isinstance(src, str):

            if media_type is None:
                media_type = MediaType.detect(src)

            with open(src, "rb") as file_obj:
                upload_result = self._upload(
                    file_obj,
                    media_type=media_type,
                    only_use_crux_domains=only_use_crux_domains,
                )

        else:
            raise TypeError("Invalid Data Type for source path: {}".format(type(src)))

        if upload_result:
            # Refresh metadata to reflect actual size after uploading the file.
            if self.refresh():
                return self
            else:
                raise CruxClientError(
                    "Error refreshing metadata for resource {id}".format(id=self.id)
                )
        else:
            raise CruxClientError(
                "Unable to upload file {file_name} to path {path}".format(
                    file_name=self.name, path=self.path
                )
            )



"""Module contains functions to create Resource objects."""

from typing import Any, Dict, Union  # noqa: F401

from crux._client import CruxClient
from crux.models.file import File
from crux.models.folder import Folder


def get_resource_object(resource_type, data, connection=None):
    # type: (str, Dict[str, Any], CruxClient) -> Union[File, Folder]
    """Creates resource object based on its type.

    Args:
        resource_type (str): Type of resource which needs to be created.
        data (dict): Dictionary which contains serialized resource data.
        connection (CruxClient): Connection Object. Defaults to None.

    Returns:
        crux.models.Resource: Resource or its Child Object.

    Raises:
        TypeError: If it is unable to detect resource type.
    """
    if resource_type == "file":
        return File.from_dict(data, connection=connection)
    elif resource_type == "folder":
        return Folder.from_dict(data, connection=connection)
    else:
        raise TypeError("Invalid Resource Type")



"""Modules contains set of utility functions."""

import logging
import posixpath
import re
from typing import List, Tuple  # noqa: F401

from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import (  # Dynamic load pylint: disable=import-error
    Retry,
)

from crux._compat import urllib_quote


DEFAULT_CHUNK_SIZE = 10485760  # 10 MB
DELIVERY_ID_REGEX = re.compile(r"^[a-zA-Z0-9]+\.[0-9]+$")
TRACE = 5


def quote(data):
    # type: (str) -> str
    """Percent Encodes the string.

    Args:
        data (str): String which is required to be Percent Encoded.

    Returns:
        str: Percent encoded string.
    """

    # With current setting, it will encode all necessary characters.
    # To prevent any character from getting percent encoded, add it to safe_characters
    # For eg: safe_characters="/%$"
    # https://docs.python.org/2/library/urllib.html
    safe_characters = ""
    return urllib_quote(data, safe=safe_characters)


def url_builder(url_base, url_prefix, url_path_list):
    # type: (str, str, List[str]) -> str
    """Builds Valid Percent Encoded URL.

    Args:
        url_base (str): Base of URL comprised of HTTP protocol scheme and Hostname.
        url_prefix (str): URL Prefix used by API Platform.
        url_path_list (:obj:`list` of :obj:`str`): List of Context Path elements.

    Returns:
        str: Valid Percent Encoded URL.
    """
    if url_prefix:
        valid_url = [url_base, quote(url_prefix)]
    else:
        valid_url = [url_base]

    valid_url_path_list = []

    for component in url_path_list:
        valid_url_path_list.append(quote(component))

    valid_url.extend(valid_url_path_list)

    return "/".join(valid_url)


def valid_chunk_size(chunk_size):
    # type: (int) -> bool
    """Checks whether chunk size is multiple of 256 KiB.

    Args:
        chunk_size (int): Input chunk_size to be validated.

    Returns:
        bool: True if chunk_size is multiple of 256 KiB, False otherwise.
    """
    return not bool(chunk_size % 262144)  # 1024*256=262144


def split_posixpath_filename_dirpath(path):
    # type: (str) -> Tuple[str, str]
    """Split a POSIX path into file name and directory path.

    Args:
        path (str): POSIX path to split.

    Returns:
        (str, str): Returns a tuple of (file name, directory path) with
            trailing slash removed from directory.
    """

    filename = posixpath.basename(path)  # type: str
    dirpath = posixpath.dirname(path)  # type: str

    if not dirpath:
        dirpath = "/"

    return filename, dirpath


def str_to_bool(string):
    # type (str) -> bool
    """Converts string to boolean value.

    Args:
        string (str): Input string.

    Returns:
        bool: True if input string is "True" or "true",
            False if input string is "False" or "false".

    Raises:
        ValueError: If input string is not in True, true,
            False, false.
    """
    if string in ("True", "true"):
        return True
    elif string in ("False", "false"):
        return False
    else:
        raise ValueError("Cannot convert {} to bool".format(string))


class Headers(dict):
    """Headers Dictionary to convert Dictionary keys to lower case."""

    def __init__(self, input_dict):
        # type(dict) -> None
        super(Headers, self).__init__()
        self.update(input_dict)

    def __setitem__(self, key, value):
        # type(str, str) -> None
        lower_key = key.lower()
        super(Headers, self).__setitem__(lower_key, value)

    def __getitem__(self, key):
        # type(str) -> Any
        return self.get(key)

    def update(self, input_dict):
        # type(dict) -> None
        for key, value in input_dict.items():
            self[key.lower()] = value

    def get(self, key):
        # type(str) -> Any
        return super(Headers, self).get(key.lower())


def get_session(session_class=Session, retries=None, proxies=None):
    # type (Type[Session], requests.packages.urllib3.util.retry.Retry, Dict) -> Session
    """Gets the session object.
    Args:
        session_class (Session): Session class. Defaults to Session.
        retries (requests.packages.urllib3.util.retry.Retry): Retry object.
        proxies (dict): Dictionary of Proxy urls.

    Returns:
        requests.Session: Session Object.

    Raises:
        TypeError: If session_class is not subclass of requests.Session.
    """
    if not issubclass(session_class, Session):
        raise TypeError("session_class should be subclass of requests.Session")

    session = session_class()

    if retries is None:
        retries = Retry(
            total=10,
            backoff_factor=1,
            connect=6,
            read=3,
            status_forcelist=[
                429,
                500,
                502,
                503,
                504,
                520,
                521,
                522,
                523,
                524,
                525,
                527,
                530,
            ],
            method_whitelist=False,
        )

    if retries:
        session.mount("http://", HTTPAdapter(max_retries=retries))
        session.mount("https://", HTTPAdapter(max_retries=retries))

    session.proxies = proxies if proxies else {}

    return session


# google.resumable_media.requests.ResumableUpload is only compatible with JSON API endpoint.
# Signed URL uses XML API Endpoint, which requires setting specific headers.
# ResumableUploadSingedSession is added to make google.resumable_media.requests.ResumableUpload
# compatible with XML API endpoint by including XML API endpoint specific headers.
class ResumableUploadSignedSession(Session):
    """Session class to support Resumable Upload."""

    def request(  # pylint: disable=arguments-differ
        self, method, url, data=None, headers=None, **kwargs
    ):
        """Implementation of Requests' request."""

        request_headers = Headers(headers.copy() if headers is not None else {})

        if self.headers:
            request_headers.update(self.headers)

        response = super(ResumableUploadSignedSession, self).request(
            method, url, data=data, headers=request_headers, **kwargs
        )

        return response


def create_logger(namespace):
    """Creates Logger with TRACE support."""
    # (str) -> logging.RootLogger
    logging.addLevelName(TRACE, "TRACE")
    logger = logging.getLogger(namespace)
    setattr(logger, "trace", lambda *args: logger.log(TRACE, *args))
    return logger



"""Module contains AbstractJob, Job, LoadJob Model."""

from crux.models.model import CruxModel


class AbstractJob(CruxModel):
    """AbstractJob Model."""


class Job(AbstractJob):
    """Job Model."""

    @property
    def job_id(self):
        """str: Gets the Job Id."""
        return self.raw_model["jobId"]

    @property
    def status(self):
        """str: Gets the Job Status."""
        return self.raw_model["status"]

    @property
    def statistics(self):
        """str: Gets the Job Statistics."""
        return self.raw_model["statistics"]


class StitchJob(AbstractJob):
    """Stitch Job Model."""

    @property
    def job_id(self):
        """str: Gets the Job Id."""
        return self.raw_model["jobId"]

    @property
    def status(self):
        """str: Gets the Job Status."""
        return self.raw_model["status"]



"""Module contains the set of crux-python's exceptions."""

from typing import Dict, Union  # noqa: F401

import requests  # pylint: disable=unused-import


class CruxAPIError(Exception):
    """Exception which should be raised when the API response expects an error."""

    def __init__(self, message):
        # type: (Dict[str, Union[str, int]]) -> None
        """
        Args:
            message (str): Human readable string describing the exception.

        Attributes:
            message (str): Human readable string describing the exception.
            status_code (int): Exception error code.
        """
        super(CruxAPIError, self).__init__(message)
        # domainV2 has "status"
        self.status_code = message.get("statusCode", message.get("status"))
        self.message = message

    def __str__(self):
        return "{message}".format(message=self.message)


class CruxClientError(Exception):
    """Exception which should be raised when the client SDK expects an error."""

    def __init__(self, message):
        # type: (str) -> None
        """
        Args:
            message (str): Human readable string describing the exception.

        Attributes:
            message (str): Human readable string describing the exception.
        """
        super(CruxClientError, self).__init__(message)
        self.message = message

    def __str__(self):
        return "{message}".format(message=self.message)


class CruxClientHTTPError(CruxClientError):
    """Exception should be raised when SDK expects any HTTP related errors."""

    def __init__(self, message, response):
        # type: (str, requests.Response) -> None
        """
        Args:
            message (str): Human readable string describing the exception.
            response (requests.Response): Response object.

        Attributes:
            message (str): Human readable string describing the exception.
            response (requests.Response): Response object.
        """
        super(CruxClientHTTPError, self).__init__(message)
        self.message = message
        self.response = response

    def __str__(self):
        return "{message}".format(message=self.message)


class CruxClientTooManyRedirects(CruxClientError):
    """Exception should be raised when SDK gets too many redirects."""

    def __str__(self):
        return "{message}".format(message=self.message)


class CruxClientConnectionError(CruxClientError):
    """Exception should be raised when SDK expects any connection related errors."""

    def __str__(self):
        return "{message}".format(message=self.message)


class CruxClientTimeout(CruxClientError):
    """Exception should be raised when SDK expects any timeout related errors."""

    def __str__(self):
        return "{message}".format(message=self.message)


class CruxResourceNotFoundError(CruxAPIError):
    """Exception which should be raised when Crux Resource is not found."""

    def __init__(self, message):
        # type: (Dict[str, Union[str, int]]) -> None
        """
        Args:
            message (str): Human readable string describing the exception.

        Attributes:
            message (str): Human readable string describing the exception.
            status_code (int): Exception error code.
        """
        super(CruxResourceNotFoundError, self).__init__(message)
        self.message = message

    def __str__(self):
        return "{message}".format(message=self.message)
